/*
 * Copyright 2000 Hans Reiser
 */

#include "includes.h"


/* compares description block with commit block. returns 0 if they differ, 1
   if they match */
int does_desc_match_commit (struct reiserfs_journal_desc *desc, 
			    struct reiserfs_journal_commit *commit) 
{
    return (get_jcommit_trans_id (commit) == get_jdesc_trans_id (desc) &&
	    get_jcommit_len (commit) == get_jdesc_len (desc));
}


/* d_bh is descriptor, return number of block where commit block of this
   transaction is to be */
unsigned long commit_expected (reiserfs_filsys_t fs, struct buffer_head * d_bh)
{
    unsigned long offset;
    struct reiserfs_journal_desc * desc;


    desc = (struct reiserfs_journal_desc *)d_bh->b_data;
    offset = d_bh->b_blocknr - get_sb_journal_1st_block (fs->s_ondisk_sb);
    return get_sb_journal_1st_block (fs->s_ondisk_sb) + 
	((offset + get_jdesc_len (desc) + 1) % get_sb_journal_size (fs->s_ondisk_sb));
}


/* d_bh contains journal descriptor, return number of block where descriptor
   block of next transaction should be */
unsigned long next_desc_expected (reiserfs_filsys_t fs, struct buffer_head * d_bh)
{
    unsigned long offset;
    struct reiserfs_journal_desc * desc;


    desc = (struct reiserfs_journal_desc *)d_bh->b_data;
    offset = d_bh->b_blocknr - get_sb_journal_1st_block (fs->s_ondisk_sb);
    return get_sb_journal_1st_block (fs->s_ondisk_sb) + 
	((offset + get_jdesc_len (desc) + 2) % get_sb_journal_size (fs->s_ondisk_sb));
}


static int is_valid_transaction (reiserfs_filsys_t fs, struct buffer_head * d_bh)
{
    struct buffer_head * c_bh;
    struct reiserfs_journal_desc * desc;
    struct reiserfs_journal_commit * commit;
    

    if (!d_bh || who_is_this (d_bh->b_data, d_bh->b_size) != THE_JDESC)
	return 0;

    /* read expected commit block and compare with descriptor block */
    c_bh = bread (fs->s_journal_dev, commit_expected (fs, d_bh), fs->s_blocksize);
    if (!c_bh)
	return 0;

    desc = (struct reiserfs_journal_desc *)d_bh->b_data;
    commit = (struct reiserfs_journal_commit *)c_bh->b_data ;
    if (does_desc_match_commit (desc, commit)) {
	brelse (c_bh);
	return 1;
    }
    brelse (c_bh);
    return 0;
}


/* read the journal and find the oldest and newest transactions */
void get_boundary_transactions (reiserfs_filsys_t fs, unsigned long * oldest,
				unsigned long * newest)
{
    struct reiserfs_super_block * sb;
    unsigned long j_cur;
    unsigned long j_start;
    unsigned long j_size;
    struct buffer_head * d_bh;
    __u32 newest_trans_id, oldest_trans_id, trans_id;
    struct reiserfs_journal_desc * j_desc;

    sb = fs->s_ondisk_sb;
    
    j_start = get_sb_journal_1st_block (fs->s_ondisk_sb);
    j_size = get_sb_journal_size (fs->s_ondisk_sb);
    
    oldest_trans_id = 0xffffffff;
    newest_trans_id = 0;

    for (j_cur = 0; j_cur < j_size; j_cur ++) {
	d_bh = bread (fs->s_journal_dev, j_start + j_cur, fs->s_blocksize);

	if (!is_valid_transaction (fs, d_bh)) {
	    brelse (d_bh);
	    continue;
	}

	j_desc = (struct reiserfs_journal_desc *)d_bh->b_data;
	
	trans_id = get_jdesc_trans_id (j_desc);
	if (trans_id < oldest_trans_id) {
	    oldest_trans_id = trans_id;
	    *oldest = d_bh->b_blocknr;
	}

	if (trans_id > newest_trans_id) {
	    newest_trans_id = trans_id;
	    *newest = d_bh->b_blocknr;
	}

	j_cur += get_jdesc_len (j_desc) + 1;
	brelse (d_bh);
    }
    return;
}



/* desc_block is descriptor block of valid transaction, return descriptor
   block of next transaction if it is a more recent one */
unsigned long next_transaction (reiserfs_filsys_t fs, unsigned long desc_block)
{
    struct buffer_head * d_bh, * next_d_bh;
    struct reiserfs_journal_desc * desc, * next_desc;
    int retval;


    retval = 0;
    d_bh = 0;
    next_d_bh = 0;

    d_bh = bread (fs->s_journal_dev, desc_block, fs->s_blocksize);
    if (!is_valid_transaction (fs, d_bh))
	goto out;

    desc_block = next_desc_expected (fs, d_bh);
    while (1) {
	next_d_bh = bread (fs->s_journal_dev, desc_block, fs->s_blocksize);
	if (is_valid_transaction (fs, next_d_bh))
	    break;
	brelse (next_d_bh);
	desc_block = get_sb_journal_1st_block (fs->s_ondisk_sb) +
	    ((desc_block + 1 - get_sb_journal_1st_block (fs->s_ondisk_sb)) %
	    get_sb_journal_size (fs->s_ondisk_sb));
    }


    desc = (struct reiserfs_journal_desc *)d_bh->b_data;
    next_desc = (struct reiserfs_journal_desc *)next_d_bh->b_data;
    if (get_jdesc_trans_id (desc) < get_jdesc_trans_id (next_desc))
	retval = next_d_bh->b_blocknr;

out:
    brelse (d_bh);
    brelse (next_d_bh);
    return retval;
}

