/*
 * Copyright 1996, 1997, 1998 Hans Reiser, see reiserfs/README for licensing and copyright details
 */

#include "includes.h"
#include <stdarg.h>
#include <limits.h>
#include <printf.h>


static int _arginfo (const struct printf_info *info, size_t n,
		     int *argtypes)
{
    if (n > 0)
	argtypes[0] = PA_POINTER;
    return 1;
}

#if 0
static int _arginfo2 (const struct printf_info *info, size_t n,
		     int *argtypes)
{
    if (n > 0)
	argtypes[0] = PA_INT;
    return 1;
}
#endif


#define FPRINTF \
    if (len == -1) {\
	return -1;\
    }\
    len = fprintf (stream, "%*s",\
		   info->left ? -info->width : info->width, buffer);\
    free (buffer);\
    return len;\


/* %b */
static int print_block_head (FILE * stream,
			     const struct printf_info *info,
			     const void *const *args)
{
    const struct buffer_head * bh;
    char * buffer;
    int len;

    bh = *((const struct buffer_head **)(args[0]));
    len = asprintf (&buffer, "level=%d, nr_items=%d, free_space=%d rdkey",
		    B_LEVEL (bh), B_NR_ITEMS (bh), B_FREE_SPACE (bh));
    FPRINTF;
}


/* %K */
static int print_short_key (FILE * stream,
			    const struct printf_info *info,
			    const void *const *args)
{
    const struct key * key;
    char * buffer;
    int len;

    key = *((const struct key **)(args[0]));
    len = asprintf (&buffer, "%u %u", get_key_dirid (key),
		    get_key_objectid (key));
    FPRINTF;
}


/* %k */
static int print_key (FILE * stream,
		      const struct printf_info *info,
		      const void *const *args)
{
    const struct key * key;
    char * buffer;
    int len;

    key = *((const struct key **)(args[0]));
    len = asprintf (&buffer, "%u %u 0x%Lx %s (%d)",  
		    get_key_dirid (key), get_key_objectid (key), get_offset (key),
		    key_of_what (key), get_type (key));
    FPRINTF;
}


/* %H */
static int print_item_head (FILE * stream,
			    const struct printf_info *info,
			    const void *const *args)
{
    const struct item_head * ih;
    char * buffer;
    int len;

    ih = *((const struct item_head **)(args[0]));
    len = asprintf (&buffer, "%u %u 0x%Lx %s (%d), "
		    "len %u, location %u entry count %u, fsck need %u, format %s",
		    get_key_dirid (&ih->ih_key), get_key_objectid (&ih->ih_key), 
		    get_offset (&ih->ih_key), key_of_what (&ih->ih_key), get_type (&ih->ih_key),
		    get_ih_item_len (ih), get_ih_location (ih), get_ih_entry_count (ih), 
		    get_ih_flags (ih),
		    get_ih_key_format (ih) == KEY_FORMAT_2 ? "new" : 
		    ((get_ih_key_format (ih) == KEY_FORMAT_1) ? "old" : "BAD"));
    FPRINTF;
}


static int print_disk_child (FILE * stream,
			     const struct printf_info *info,
			     const void *const *args)
{
    const struct disk_child * dc;
    char * buffer;
    int len;

    dc = *((const struct disk_child **)(args[0]));
    len = asprintf (&buffer, "[dc_number=%u, dc_size=%u]", le32_to_cpu (dc->dc_block_number),
		    le16_to_cpu (dc->dc_size));
    FPRINTF;
}


char ftypelet (mode_t mode)
{
    if (S_ISBLK (mode))
	return 'b';
    if (S_ISCHR (mode))
	return 'c';
    if (S_ISDIR (mode))
	return 'd';
    if (S_ISREG (mode))
	return '-';
    if (S_ISFIFO (mode))
	return 'p';
    if (S_ISLNK (mode))
	return 'l';
    if (S_ISSOCK (mode))
	return 's';
    return '?';
}


static int rwx (FILE * stream, mode_t mode)
{
    return fprintf (stream, "%c%c%c",
		    (mode & S_IRUSR) ? 'r' : '-',
		    (mode & S_IWUSR) ? 'w' : '-',
		    (mode & S_IXUSR) ? 'x' : '-');
}


/* %M */
static int print_sd_mode (FILE * stream,
			  const struct printf_info *info,
			  const void *const *args)
{
    int len = 0;
    mode_t mode;

    mode = *(mode_t *)args[0];
    len = fprintf (stream, "%c", ftypelet (mode));
    len += rwx (stream, (mode & 0700) << 0);
    len += rwx (stream, (mode & 0070) << 3);
    len += rwx (stream, (mode & 0007) << 6);
    return len;
}



void reiserfs_warning (FILE * fp, const char * fmt, ...)
{
    static int registered = 0;
    va_list args;

    if (!registered) {
	registered = 1;
	
	register_printf_function ('K', print_short_key, _arginfo);
	register_printf_function ('k', print_key, _arginfo);
	register_printf_function ('H', print_item_head, _arginfo);
	register_printf_function ('b', print_block_head, _arginfo);
	register_printf_function ('y', print_disk_child, _arginfo);
	register_printf_function ('M', print_sd_mode, _arginfo);
    }

    va_start (args, fmt);
    vfprintf (fp, fmt, args);
    va_end (args);
}


static char * vi_type (struct virtual_item * vi)
{
    static char *types[]={"directory", "direct", "indirect", "stat data"};

    if (vi->vi_type & VI_TYPE_STAT_DATA)
	return types[3];
    if (vi->vi_type & VI_TYPE_INDIRECT)
	return types[2];
    if (vi->vi_type & VI_TYPE_DIRECT)
	return types[1];
    if (vi->vi_type & VI_TYPE_DIRECTORY)
	return types[0];

    reiserfs_panic ("vi_type: 6000: unknown type (0x%x)", vi->vi_type);
    return NULL;
}


void print_virtual_node (struct virtual_node * vn)
{
    int i, j;
  
    printf ("VIRTUAL NODE CONTAINS %d items, has size %d,%s,%s, ITEM_POS=%d POS_IN_ITEM=%d MODE=\'%c\'\n",
	    vn->vn_nr_item, vn->vn_size,
	    (vn->vn_vi[0].vi_type & VI_TYPE_LEFT_MERGEABLE )? "left mergeable" : "", 
	    (vn->vn_vi[vn->vn_nr_item - 1].vi_type & VI_TYPE_RIGHT_MERGEABLE) ? "right mergeable" : "",
	    vn->vn_affected_item_num, vn->vn_pos_in_item, vn->vn_mode);


    for (i = 0; i < vn->vn_nr_item; i ++) {
	printf ("%s %d %d", vi_type (&vn->vn_vi[i]), i, vn->vn_vi[i].vi_item_len);
	if (vn->vn_vi[i].vi_entry_sizes)
	{
	    printf ("It is directory with %d entries: ", vn->vn_vi[i].vi_entry_count);
	    for (j = 0; j < vn->vn_vi[i].vi_entry_count; j ++)
		printf ("%d ", vn->vn_vi[i].vi_entry_sizes[j]);
	}
	printf ("\n");
    }
}


void print_path (struct tree_balance * tb, struct path * path)
{
    int offset = path->path_length;
    struct buffer_head * bh;

    printf ("Offset    Bh     (b_blocknr, b_count) Position Nr_item\n");
    while ( offset > ILLEGAL_PATH_ELEMENT_OFFSET ) {
	bh = PATH_OFFSET_PBUFFER (path, offset);
	printf ("%6d %10p (%9lu, %7d) %8d %7d\n", offset, 
		bh, bh ? bh->b_blocknr : 0, bh ? bh->b_count : 0,
		PATH_OFFSET_POSITION (path, offset), bh ? B_NR_ITEMS (bh) : -1);
	
	offset --;
    }
}



void print_directory_item (FILE * fp, reiserfs_filsys_t fs,
			   struct buffer_head * bh, struct item_head * ih)
{
    int i;
    int namelen;
    struct reiserfs_de_head * deh;
    char * name;
/*    static char namebuf [80];*/

    if (!I_IS_DIRECTORY_ITEM (ih))
	return;

    //printk ("\n%2%-25s%-30s%-15s%-15s%-15s\n", "    Name", "length", "Object key", "Hash", "Gen number", "Status");
    reiserfs_warning (fp, "%3s: %-25s%s%-22s%-12s%s\n", "###", "Name", "length", "    Object key", "   Hash", "Gen number");
    deh = B_I_DEH (bh, ih);
    for (i = 0; i < get_ih_entry_count (ih); i ++, deh ++) {
	if (dir_entry_bad_location (deh, ih, i == 0 ? 1 : 0)) {
	    reiserfs_warning (fp, "%3d: wrong entry location %u, deh_offset %u\n",
			      i, deh_location (deh), deh_offset (deh));
	    continue;
	}
	if (i && dir_entry_bad_location (deh - 1, ih, ((i - 1) == 0) ? 1 : 0))
	    /* previous entry has bad location so we can not calculate entry
               length */
	    namelen = 25;
	else
	    namelen = name_length (ih, deh, i);

	name = name_in_entry (deh, i);
	reiserfs_warning (fp, "%3d: \"%-25.*s\"(%3d)%20K%12d%5d, loc %u, state %x %s\n", 
			  i, namelen, name, namelen,
			  (struct key *)&(deh->deh_dir_id),
			  GET_HASH_VALUE (deh->deh_offset), GET_GENERATION_NUMBER (deh->deh_offset),
			  deh_location (deh), deh->deh_state,
			  code2name (find_hash_in_use (name, namelen, 
						       GET_HASH_VALUE (deh->deh_offset),
						       fs ? get_sb_hash_code (fs->s_ondisk_sb) : UNSET_HASH)));
	/*fs ? (is_properly_hashed (fs, name, namelen, deh_offset (deh)) ? "" : "(BROKEN)") : "??");*/
    }
}


//
// printing of indirect item
//
static void start_new_sequence (__u32 * start, int * len, __u32 new)
{
    *start = new;
    *len = 1;
}


static int sequence_finished (__u32 start, int * len, __u32 new)
{
    if (start == INT_MAX)
	return 1;

    if (start == 0 && new == 0) {
	(*len) ++;
	return 0;
    }
    if (start != 0 && (start + *len) == new) {
	(*len) ++;
	return 0;
    }
    return 1;
}

static void print_sequence (FILE * fp, __u32 start, int len)
{
    if (start == INT_MAX)
	return;

    if (len == 1)
	reiserfs_warning (fp, " %d", start);
    else
	reiserfs_warning (fp, " %d(%d)", start, len);
}


void print_indirect_item (FILE * fp, struct buffer_head * bh, int item_num)
{
    struct item_head * ih;
    int j;
    __u32 * unp, prev = INT_MAX;
    int num;

    ih = B_N_PITEM_HEAD (bh, item_num);
    unp = (__u32 *)B_I_PITEM (bh, ih);

    if (get_ih_item_len (ih) % UNFM_P_SIZE)
	reiserfs_warning (fp, "print_indirect_item: invalid item len");  

    reiserfs_warning (fp, "%d pointers\n[ ", I_UNFM_NUM (ih));
    for (j = 0; j < I_UNFM_NUM (ih); j ++) {
	if (sequence_finished (prev, &num, unp[j])) {
	    print_sequence (fp, prev, num);
	    start_new_sequence (&prev, &num, unp[j]);
	}
    }
    print_sequence (fp, prev, num);
    reiserfs_warning (fp, "]\n");
}


char timebuf[256];

char * timestamp (time_t t)
{
    strftime (timebuf, 256, "%m/%d/%Y %T", localtime (&t));
    return timebuf;
}

static int print_stat_data (FILE * fp, struct buffer_head * bh, struct item_head * ih, int alltimes)
{
    struct stat_data * sd = (struct stat_data *)B_I_PITEM (bh, ih);
    struct stat_data_v1 * sd_v1 = (struct stat_data_v1 *)B_I_PITEM (bh, ih);
    int retval;
    

    /* we can not figure out whether it is new stat data or old by key_format
       macro. Stat data's key looks identical in both formats */
    if (get_ih_key_format (ih) == KEY_FORMAT_1) {
	reiserfs_warning (fp, "(OLD SD), mode %M, size %u, nlink %u, uid %d, FDB %d, mtime %s blocks %d", 
		sd_v1->sd_mode, sd_v1->sd_size, sd_v1->sd_nlink, sd_v1->sd_uid, 
		sd_v1->sd_first_direct_byte, timestamp (sd_v1->sd_mtime), sd_v1->u.sd_blocks);
	retval = (S_ISLNK (sd_v1->sd_mode)) ? 1 : 0;
    } else {
	reiserfs_warning (fp, "(NEW SD), mode %M, size %Lu, nlink %u, mtime %s blocks %d, uid %d", 
		sd->sd_mode, sd->sd_size, sd->sd_nlink,
		timestamp (sd->sd_mtime), sd->sd_blocks, sd->sd_uid);
	retval = (S_ISLNK (sd->sd_mode)) ? 1 : 0;
    }

    if (alltimes)
	reiserfs_warning (fp, "%s %s\n", timestamp (sd->sd_ctime), timestamp (sd->sd_atime));
    reiserfs_warning (fp, "\n");
    return retval;
}


/* used by debugreiserfs/scan.c */
void reiserfs_print_item (FILE * fp, struct buffer_head * bh,
			  struct item_head * ih)
{
    reiserfs_warning (stdout, "block %lu, item %d: %H\n",
		      bh->b_blocknr, ih - B_N_PITEM_HEAD (bh, 0), ih);
    if (is_stat_data_ih (ih)) {
	print_stat_data (fp, bh, ih, 0/*all times*/);
	return;
    }
    if (is_indirect_ih (ih)) {
	print_indirect_item (fp, bh, ih - B_N_PITEM_HEAD (bh, 0));
	return;
    }
    if (is_direct_ih (ih)) {
	reiserfs_warning (fp, "direct item: block %lu, start %d, %d bytes\n",
			  bh->b_blocknr, get_ih_location (ih), get_ih_item_len (ih));
	return;
    }

    print_directory_item (fp, 0, bh, ih);    
}


/* this prints internal nodes (4 keys/items in line) (dc_number,
   dc_size)[k_dirid, k_objectid, k_offset, k_uniqueness](dc_number,
   dc_size)...*/
static int print_internal (FILE * fp, struct buffer_head * bh, int first, int last)
{
    struct key * key;
    struct disk_child * dc;
    int i;
    int from, to;

    if (!is_internal_node (bh))
	return 1;

    if (first == -1) {
	from = 0;
	to = B_NR_ITEMS (bh);
    } else {
	from = first;
	to = last < B_NR_ITEMS (bh) ? last : B_NR_ITEMS (bh);
    }

    reiserfs_warning (fp, "INTERNAL NODE (%ld) contains %b\n",  bh->b_blocknr, bh);

    dc = B_N_CHILD (bh, from);
    reiserfs_warning (fp, "PTR %d: %y ", from, dc);

    for (i = from, key = B_N_PDELIM_KEY (bh, from), dc ++; i < to; i ++, key ++, dc ++) {
	reiserfs_warning (fp, "KEY %d: %20k PTR %d: %20y ", i, key, i + 1, dc);
	if (i && i % 4 == 0)
	    reiserfs_warning (fp, "\n");
    }
    reiserfs_warning (fp, "\n");
    return 0;
}



static int is_symlink = 0;
static int print_leaf (FILE * fp, reiserfs_filsys_t fs, struct buffer_head * bh,
		       int print_mode, int first, int last)
{
    struct block_head * blkh;
    struct item_head * ih;
    int i;
    int from, to;
    int nr;

    nr = does_node_have_ih_array (bh->b_data, bh->b_size);
    if (!nr)
	return 1;
    

    blkh = B_BLK_HEAD (bh);
    ih = B_N_PITEM_HEAD (bh,0);

    reiserfs_warning (fp,
		      "\n===================================================================\n");
    reiserfs_warning (fp, "LEAF NODE (%ld) contains %b (real items %d)\n",
		      bh->b_blocknr, bh, nr);

    if (!(print_mode & PRINT_LEAF_ITEMS)) {
	reiserfs_warning (fp, "FIRST ITEM_KEY: %k, LAST ITEM KEY: %k\n",
			   &(ih->ih_key), &((ih + get_blkh_nr_items (blkh) - 1)->ih_key));
	return 0;
    }

    if (first < 0 || first > nr - 1) 
	from = 0;
    else 
	from = first;

    if (last < 0 || last > nr)
	to = nr;
    else
	to = last;


    reiserfs_warning (fp,
		       "-------------------------------------------------------------------------------\n"
		       "|###|type|ilen|f/sp| loc|fmt|fsck|                   key                      |\n"
		       "|   |    |    |e/cn|    |   |need|                                            |\n");
    for (i = from; i < to; i++) {
	reiserfs_warning (fp,
			  "-------------------------------------------------------------------------------\n"
			  "|%3d|%30H|\n", i, ih + i);

	if (I_IS_STAT_DATA_ITEM(ih+i)) {
	    is_symlink = print_stat_data (fp, bh, ih + i, 0/*all times*/);
	    continue;
	}

	if (I_IS_DIRECTORY_ITEM(ih+i)) {
	    print_directory_item (fp, fs, bh, ih+i);
	    continue;
	}

	if (I_IS_INDIRECT_ITEM(ih+i)) {
	    print_indirect_item (fp, bh, i);
	    continue;
	}

	if (I_IS_DIRECT_ITEM(ih+i)) {
	    int j = 0;
	    if (is_symlink || print_mode & PRINT_DIRECT_ITEMS) {
		reiserfs_warning (fp, "\"");
		while (j < get_ih_item_len (&ih[i])) {
		    if (B_I_PITEM(bh,ih+i)[j] == 10)
			reiserfs_warning (fp, "\\n");
		    else
			reiserfs_warning (fp, "%c", B_I_PITEM(bh,ih+i)[j]);
		    j ++;
		}
		reiserfs_warning (fp, "\"\n");
	    }
	    continue;
	}
    }
    reiserfs_warning (fp, "===================================================================\n");
    return 0;
}



/* return 1 if this is not super block */
static int print_super_block (FILE * fp, struct buffer_head * bh)
{
    struct reiserfs_super_block * sb = (struct reiserfs_super_block *)(bh->b_data);
    int skipped, data_blocks;
    
    if (is_reiser2fs_magic_string (sb))
	reiserfs_warning (fp, "Super block of format 3.6 found on the 0x%x in block %ld\n", 
			  bh->b_dev, bh->b_blocknr);
    else if (is_reiserfs_magic_string (sb))
	reiserfs_warning (fp, "Super block of format 3.5 found on the 0x%x in block %ld\n",
			  bh->b_dev, bh->b_blocknr);
    else if (is_prejournaled_reiserfs (sb)) {
	reiserfs_warning (fp, "Prejournaled reiserfs super block found. Not supported here. Use proper tools instead\n");
	return 1;
    } else
	// no reiserfs signature found in the block
	return 1;

    reiserfs_warning (fp, "Block count %u\n", get_sb_block_count (sb));
    reiserfs_warning (fp, "Blocksize %d\n", get_sb_block_size (sb));
    reiserfs_warning (fp, "Free blocks %u\n", get_sb_free_blocks (sb));
    skipped = bh->b_blocknr;
    data_blocks = get_sb_block_count (sb) - skipped - 1 -
	get_sb_bmap_nr (sb) - (get_sb_journal_dev (sb) ? 0 : get_sb_journal_size (sb) + 1) - get_sb_free_blocks (sb);
    if (get_sb_journal_dev (sb))
	reiserfs_warning (fp, "Busy blocks (skipped %d, bitmaps - %d,\n"
			  "1 super block, %d data blocks\n", 
			  skipped, get_sb_bmap_nr (sb), data_blocks);
    else
	reiserfs_warning (fp, "Busy blocks (skipped %d, bitmaps - %d, journal blocks - %d\n"
			  "1 super block, %d data blocks\n", 
			  skipped, get_sb_bmap_nr (sb), get_sb_journal_size (sb) + 1 , data_blocks);
    
    reiserfs_warning (fp, "Root block %u\n", get_sb_root_block (sb));
    reiserfs_warning (fp, "Journal block (first) %d\n", get_sb_journal_1st_block (sb));
    if (get_sb_journal_dev (sb))
	reiserfs_warning (fp, "Journal device %x\n", get_sb_journal_dev (sb));    
    reiserfs_warning (fp, "Journal orig size %d\n", get_sb_journal_size (sb));
    reiserfs_warning (fp, "Filesystem state %s\n",
		      (get_sb_state (sb) == REISERFS_VALID_FS) ? "VALID" : "ERROR");
 
#if 0
    __u32 s_journal_trans_max ;           /* max number of blocks in a transaction.  */
    __u32 s_journal_block_count ;         /* total size of the journal. can change over time  */
    __u32 s_journal_max_batch ;           /* max number of blocks to batch into a trans */
    __u32 s_journal_max_commit_age ;      /* in seconds, how old can an async commit be */
    __u32 s_journal_max_trans_age ;       /* in seconds, how old can a transaction be */
#endif

    reiserfs_warning (fp, "Tree height %d\n", get_sb_tree_height (sb));
    reiserfs_warning (fp, "Hash function used to sort names: %s\n",
		      code2name (get_sb_hash_code (sb)));
    reiserfs_warning (fp, "Objectid map size %d, max %d\n", get_sb_oid_cursize (sb),
		      get_sb_oid_maxsize (sb));
    reiserfs_warning (fp, "Filesystem format \"%s\"\n",
		      get_sb_version (sb) == REISERFS_VERSION_1 ? "3.5" : "3.6(or converted)");
    switch (get_sb_fsck_state (sb)) {
    case 0:
	break;
    case PASS_0_DONE:
	reiserfs_warning (fp, "Fsck state: pass 0 done\n");
	break;
    case PASS_1_DONE:
	reiserfs_warning (fp, "Fsck state: pass 1 done\n");
	break;
    case TREE_IS_BUILT :
	reiserfs_warning (fp, "Fsck state: tree is built\n");
	break;
    default:
	reiserfs_warning (fp, "Fsck state: unknown\n");
	break;
    }
    return 0;
}


static int print_desc_block (FILE * fp, struct buffer_head * bh)
{
    struct reiserfs_journal_desc * desc;

    desc = (struct reiserfs_journal_desc *)(bh->b_data);

    if (memcmp(desc->j_magic, JOURNAL_DESC_MAGIC, 8))
	return 1;

    reiserfs_warning (fp, "Desc block %lu (j_trans_id %ld, j_mount_id %ld, j_len %ld)\n",
		       bh->b_blocknr, get_jdesc_trans_id (desc),
		      get_jdesc_mount_id (desc), get_jdesc_len (desc));

    return 0;
}


void print_block (FILE * fp, reiserfs_filsys_t fs, 
		  struct buffer_head * bh, ...)//int print_mode, int first, int last)
{
    va_list args;
    int mode, first, last;
    
    va_start (args, bh);

    if ( ! bh ) {
	reiserfs_warning (stderr, "print_block: buffer is NULL\n");
	return;
    }

    mode = va_arg (args, int);
    first = va_arg (args, int);
    last = va_arg (args, int);
    if (print_desc_block (fp, bh))
	if (print_super_block (fp, bh))
	    if (print_leaf (fp, fs, bh, mode, first, last))
		if (print_internal (fp, bh, first, last))
		    reiserfs_warning (fp, "Block %ld contains unformatted data\n", bh->b_blocknr);
}


void print_tb (int mode, int item_pos, int pos_in_item, struct tree_balance * tb, char * mes)
{
  int h = 0;
  int i;
  struct buffer_head * tbSh, * tbFh;


  if (!tb)
    return;

  printf ("\n********************** PRINT_TB for %s *******************\n", mes);
  printf ("MODE=%c, ITEM_POS=%d POS_IN_ITEM=%d\n", mode, item_pos, pos_in_item);
  printf ("*********************************************************************\n");

  printf ("* h *    S    *    L    *    R    *   F   *   FL  *   FR  *  CFL  *  CFR  *\n");
/*
01234567890123456789012345678901234567890123456789012345678901234567890123456789
       1        2         3         4         5         6         7         8
  printk ("*********************************************************************\n");
*/
  
  
  for (h = 0; h < sizeof(tb->insert_size) / sizeof (tb->insert_size[0]); h ++) {
    if (PATH_H_PATH_OFFSET (tb->tb_path, h) <= tb->tb_path->path_length && 
	PATH_H_PATH_OFFSET (tb->tb_path, h) > ILLEGAL_PATH_ELEMENT_OFFSET) {
      tbSh = PATH_H_PBUFFER (tb->tb_path, h);
      tbFh = PATH_H_PPARENT (tb->tb_path, h);
    } else {
      /*      printk ("print_tb: h=%d, PATH_H_PATH_OFFSET=%d, path_length=%d\n", 
	      h, PATH_H_PATH_OFFSET (tb->tb_path, h), tb->tb_path->path_length);*/
      tbSh = 0;
      tbFh = 0;
    }
    printf ("* %d * %3ld(%2d) * %3ld(%2d) * %3ld(%2d) * %5ld * %5ld * %5ld * %5ld * %5ld *\n",
	    h, 
	    (tbSh) ? (tbSh->b_blocknr):(-1),
	    (tbSh) ? tbSh->b_count : -1,
	    (tb->L[h]) ? (tb->L[h]->b_blocknr):(-1),
	    (tb->L[h]) ? tb->L[h]->b_count : -1,
	    (tb->R[h]) ? (tb->R[h]->b_blocknr):(-1),
	    (tb->R[h]) ? tb->R[h]->b_count : -1,
	    (tbFh) ? (tbFh->b_blocknr):(-1),
	    (tb->FL[h]) ? (tb->FL[h]->b_blocknr):(-1),
	    (tb->FR[h]) ? (tb->FR[h]->b_blocknr):(-1),
	    (tb->CFL[h]) ? (tb->CFL[h]->b_blocknr):(-1),
	    (tb->CFR[h]) ? (tb->CFR[h]->b_blocknr):(-1));
  }

  printf ("*********************************************************************\n");


  /* print balance parameters for leaf level */
  h = 0;
  printf ("* h * size * ln * lb * rn * rb * blkn * s0 * s1 * s1b * s2 * s2b * curb * lk * rk *\n");
  printf ("* %d * %4d * %2d * %2d * %2d * %2d * %4d * %2d * %2d * %3d * %2d * %3d * %4d * %2d * %2d *\n",
	  h, tb->insert_size[h], tb->lnum[h], tb->lbytes, tb->rnum[h],tb->rbytes, tb->blknum[h], 
	  tb->s0num, tb->s1num,tb->s1bytes,  tb->s2num, tb->s2bytes, tb->cur_blknum, tb->lkey[h], tb->rkey[h]);


/* this prints balance parameters for non-leaf levels */
  do {
    h++;
    printf ("* %d * %4d * %2d *    * %2d *    * %2d *\n",
    h, tb->insert_size[h], tb->lnum[h], tb->rnum[h], tb->blknum[h]);
  } while (tb->insert_size[h]);

  printf ("*********************************************************************\n");


  /* print FEB list (list of buffers in form (bh (b_blocknr, b_count), that will be used for new nodes) */
  h = 0;
  for (i = 0; i < sizeof (tb->FEB) / sizeof (tb->FEB[0]); i ++)
    printf ("%s%p (%lu %d)", i == 0 ? "FEB list: " : ", ", tb->FEB[i], tb->FEB[i] ? tb->FEB[i]->b_blocknr : 0,
	    tb->FEB[i] ? tb->FEB[i]->b_count : 0);
  printf ("\n");

  printf ("********************** END OF PRINT_TB *******************\n\n");

}


static void print_bmap_block (FILE * fp, int i, struct buffer_head * bmap, int blocks, int silent)
{
    int j, k;
    int bits = bmap->b_size * 8;
    int zeros = 0, ones = 0;
  
    reiserfs_warning (fp, "#%d: block %lu: ", i, bmap->b_blocknr);

    if (test_bit (0, bmap->b_data)) {
	/* first block addressed by this bitmap block is used */
	ones ++;
	if (!silent)
	    reiserfs_warning (fp, "Busy (%d-", i * bits);
	for (j = 1; j < blocks; j ++) {
	    while (test_bit (j, bmap->b_data)) {
		ones ++;
		if (j == blocks - 1) {
		    if (!silent)
			reiserfs_warning (fp, "%d)\n", j + i * bits);
		    goto end;
		}
		j++;
	    }
	    if (!silent)
		reiserfs_warning (fp, "%d) Free(%d-", j - 1 + i * bits, j + i * bits);

	    while (!test_bit (j, bmap->b_data)) {
		zeros ++;
		if (j == blocks - 1) {
		    if (!silent)
			reiserfs_warning (fp, "%d)\n", j + i * bits);
		    goto end;
		}
		j++;
	    }
	    if (!silent)
		reiserfs_warning (fp, "%d) Busy(%d-", j - 1 + i * bits, j + i * bits);

	    j --;
	end:
	}
    } else {
	/* first block addressed by this bitmap is free */
	zeros ++;
	if (!silent)
	    reiserfs_warning (fp, "Free (%d-", i * bits);
	for (j = 1; j < blocks; j ++) {
	    k = 0;
	    while (!test_bit (j, bmap->b_data)) {
		k ++;
		if (j == blocks - 1) {
		    if (!silent)
			reiserfs_warning (fp, "%d)\n", j + i * bits);
		    zeros += k;
		    goto end2;
		}
		j++;
	    }
	    zeros += k;
	    if (!silent)
		reiserfs_warning (fp, "%d) Busy(%d-", j - 1 + i * bits, j + i * bits);
	    
	    k = 0;
	    while (test_bit (j, bmap->b_data)) {
		ones ++;
		if (j == blocks - 1) {
		    if (!silent)
			reiserfs_warning (fp, "%d)\n", j + i * bits);
		    ones += k;
		    goto end2;
		}
		j++;
	    }
	    ones += k;
	    if (!silent)
		reiserfs_warning (fp, "%d) Free(%d-", j - 1 + i * bits, j + i * bits);
	
	    j --;
	end2:
	}
    }

    reiserfs_warning (fp, "used %d, free %d\n", ones, zeros);
}


/* if silent == 1, do not print details */
void print_bmap (FILE * fp, reiserfs_filsys_t fs, int silent)
{
    struct reiserfs_super_block * sb;
    int bmap_nr;
    int i;
    int bits_per_block;
    int blocks;

    sb = fs->s_ondisk_sb;
    bmap_nr = get_sb_bmap_nr (sb);
    bits_per_block = fs->s_blocksize * 8;
    blocks = bits_per_block;

    reiserfs_warning (fp, "Bitmap blocks are:\n");
    for (i = 0; i < bmap_nr; i ++) {
	if (i == bmap_nr - 1)
	    if (get_sb_block_count (sb) % bits_per_block)
		blocks = get_sb_block_count (sb) % bits_per_block;
	print_bmap_block (fp, i, SB_AP_BITMAP(fs)[i], blocks, silent);
    }

    /* check unused part of last bitmap */
    {
	int bad_unused_bitmap = 0;
	int ones;

	ones = bits_per_block - (get_sb_block_count (sb) % bits_per_block);
	if (ones == bits_per_block)
	    ones = 0;
      
	for (i = bits_per_block; --i >= blocks; )
	    if (!test_bit (i, SB_AP_BITMAP (fs)[bmap_nr - 1]->b_data))
		bad_unused_bitmap ++;

	if (bad_unused_bitmap) {
	    reiserfs_warning (fp, "Unused part of bitmap is wrong: should be %d ones, found %d zeros\n",
			       ones, bad_unused_bitmap);
	}
    }
    
}



void print_objectid_map (FILE * fp, reiserfs_filsys_t fs)
{
    int i;
    struct reiserfs_super_block * sb;
    __u32 * omap;

    sb = fs->s_ondisk_sb;
    if (fs->s_version == REISERFS_VERSION_2)
	omap = (__u32 *)(sb + 1);
    else if (fs->s_version == REISERFS_VERSION_1)
	omap = (__u32 *)((struct reiserfs_super_block_v1 *)sb + 1);
    else {
	reiserfs_warning (fp, "print_objectid_map: proper signature is not found\n");
	return;
    }
	
    reiserfs_warning (fp, "Map of objectids (super block size %d)\n",
		      (char *)omap - (char *)sb);
      
    for (i = 0; i < get_sb_oid_cursize (sb); i ++) {
	if (i % 2 == 0)
	    reiserfs_warning (fp, "busy(%u-%u) ", le32_to_cpu (omap[i]),
			      le32_to_cpu (omap[i+1]) - 1); 
	else
	    reiserfs_warning (fp, "free(%u-%u) ", 
			      le32_to_cpu (omap[i]),
			      ((i+1) == get_sb_oid_cursize (sb)) ? -1 : (le32_to_cpu (omap[i+1]) - 1));
    }

    reiserfs_warning (fp, "\nObject id array has size %d (max %d):", 
		      get_sb_oid_cursize (sb), get_sb_oid_maxsize (sb));
  
    for (i = 0; i < get_sb_oid_cursize (sb); i ++)
	reiserfs_warning (fp, "%s%u ", i % 2 ? "" : "*", le32_to_cpu (omap[i])); 
    reiserfs_warning (fp, "\n");

}


static void print_journal_header (reiserfs_filsys_t fs)
{
    struct buffer_head * bh;
    struct reiserfs_journal_header * j_head;


    bh = bread (fs->s_dev, get_sb_journal_1st_block (fs->s_ondisk_sb) +
		get_sb_journal_size (fs->s_ondisk_sb),
		fs->s_blocksize);
    if (!bh) {
	reiserfs_warning (stdout, "Could not read journal header\n");
	return;
    }

    j_head = (struct reiserfs_journal_header *)(bh->b_data);
 
    reiserfs_warning (stdout, "#######################\nJournal header (block #%lu):\n"
		      "j_last_flush_trans_id %ld\n"
		      "j_first_unflushed_offset %ld\n"
		      "j_mount_id %ld\n", 
		      bh->b_blocknr,
		      j_head->j_last_flush_trans_id,
		      j_head->j_first_unflushed_offset,
		      j_head->j_mount_id);
    brelse (bh);
}


static void print_one_transaction (reiserfs_filsys_t fs, unsigned long desc_block)
{
    struct buffer_head * d_bh, * c_bh;
    struct reiserfs_journal_desc * desc;


    d_bh = bread (fs->s_dev, desc_block, fs->s_blocksize);
    if (!d_bh) {
	reiserfs_warning (stdout, "reading descriptor block %lu failed\n",
			  desc_block);
	return;
    }

    c_bh = bread (fs->s_dev, commit_expected (fs, d_bh), fs->s_blocksize);
    if (!c_bh) {
	reiserfs_warning (stdout, "reading commit block %lu failed\n",
			  commit_expected (fs, d_bh));
	brelse (d_bh);
	return;
    }

    desc = (struct reiserfs_journal_desc *)(d_bh->b_data);
    reiserfs_warning (stdout, "Mountid %u, transid %u, desc %lu, length %u, commit %lu\n",
		      get_jdesc_mount_id (desc), get_jdesc_trans_id (desc),
		      d_bh->b_blocknr,
		      get_jdesc_len (desc), c_bh->b_blocknr);
   
    brelse (d_bh);
    brelse (c_bh);
}


/* print all valid transactions and found dec blocks */
void print_journal (reiserfs_filsys_t fs)
{
    unsigned long oldest, newest, trans;

    
    print_journal_header (fs);

    get_boundary_transactions (fs, &oldest, &newest);
    reiserfs_warning (stdout, "Oldest trans %lu, newest %lu\n",
		      oldest, newest);
    trans = oldest;
    while (trans) {
	print_one_transaction (fs, trans);
	trans = next_transaction (fs, trans);
    }    
}

#if 0
/* print all valid transactions and found dec blocks */
void print_journal_2 (reiserfs_filsys_t fs)
{
    struct reiserfs_super_block * sb;
    struct buffer_head * d_bh, * c_bh;
    struct reiserfs_journal_desc * desc ;
    struct reiserfs_journal_commit *commit ;
    int end_journal;
    int start_journal;
    int i, j;
    int first_desc_block = 0;
    int wrapped = 0;
    int valid_transactions = 0;

    sb = fs->s_ondisk_sb;
    start_journal = get_sb_journal_1st_block (sb);
    end_journal = start_journal + get_sb_journal_size (sb);
    reiserfs_warning (stdout, "Start scanning from %d\n", start_journal);

    d_bh = 0;
    desc = 0;
    for (i = start_journal; i < end_journal; i ++) {
	d_bh = bread (fs->s_dev, i, fs->s_blocksize);
	if (who_is_this (d_bh->b_data, d_bh->b_size) == THE_JDESC) {
	    int commit_block;

	    if (first_desc_block == 0)
		/* store where first desc block found */
		first_desc_block = i;

	    print_block (stdout, fs, d_bh); /* reiserfs_journal_desc structure will be printed */
	    desc = (struct reiserfs_journal_desc *)(d_bh->b_data);

	    commit_block = d_bh->b_blocknr + get_jdesc_len (desc) + 1;
	    if (commit_block >= end_journal) {
		reiserfs_warning (stdout, "-- wrapped?");
		wrapped = 1;
		break;
	    }

	    c_bh = bread (fs->s_dev, commit_block, fs->s_blocksize);
	    commit = bh_commit (c_bh);
	    if (does_desc_match_commit (desc, commit)) {
		reiserfs_warning (stdout, "commit block %d (trans_id %ld, j_len %ld) does not match\n", commit_block,
				  commit->j_trans_id, commit->j_len);
		brelse (c_bh) ;
		brelse (d_bh);
		continue;
	    }

	    valid_transactions ++;
	    reiserfs_warning (stdout, "(commit block %d) - logged blocks (", commit_block);
#if 1
	    for (j = 0; j < desc->j_len; j ++) {
		unsigned long block;

		if (j < JOURNAL_TRANS_HALF)
		    block = le32_to_cpu (desc->j_realblock[j]);
		else
		    block = le32_to_cpu (commit->j_realblock[i - JOURNAL_TRANS_HALF]);
			
		if (not_journalable (fs, block))
		    reiserfs_warning (stdout, " xxx%ldxxx", block);
		else {
		    reiserfs_warning (stdout, " %ld", block);
		    if (block_of_bitmap (fs, block))
			reiserfs_warning (stdout, "(bmp)");
		}
		if (j && (j + 1) % 10 == 0)
		    reiserfs_warning (stdout, "\n");
	    }
#endif
	    reiserfs_warning (stdout, ")\n");
	    i += desc->j_len + 1;
	    brelse (c_bh);
	}
	brelse (d_bh);
    }
    
    if (wrapped) {
	c_bh = bread (fs->s_dev, first_desc_block - 1, fs->s_blocksize);
	commit = bh_commit (c_bh);
	if (does_desc_match_commit (desc, commit)) {
	    reiserfs_warning (stdout, "No! commit block %d (trans_id %ld, j_len %ld) does not match\n",
			       first_desc_block - 1, commit->j_trans_id, commit->j_len);
	} else {
	    reiserfs_warning (stdout, "Yes! (commit block %d) - logged blocks (\n", first_desc_block - 1);
#if 1
	    for (j = 0; j < desc->j_len; j ++) {
		unsigned long block;

		if (j < JOURNAL_TRANS_HALF)
		    block = le32_to_cpu (desc->j_realblock[j]);
		else
		    block = le32_to_cpu (commit->j_realblock[i - JOURNAL_TRANS_HALF]);
			
		if (not_journalable (fs, block))
		    reiserfs_warning (stdout, " xxx%ldxxx", block);
		else {
		    reiserfs_warning (stdout, " %ld", block);
		    if (block_of_bitmap (fs, block))
			reiserfs_warning (stdout, "(bmp)");
		}
	    }
#endif
	    reiserfs_warning (stdout, "\n");
	}
	brelse (c_bh) ;
	brelse (d_bh);
    }

    reiserfs_warning (stdout, "%d valid transactions found\n", valid_transactions);

    {
	struct buffer_head * bh;
	struct reiserfs_journal_header * j_head;

	bh = bread (fs->s_dev, end_journal, fs->s_blocksize);
	j_head = (struct reiserfs_journal_header *)(bh->b_data);

	reiserfs_warning (stdout, "#######################\nJournal header (block #%lu):\n"
			  "j_last_flush_trans_id %ld\n"
			  "j_first_unflushed_offset %ld\n"
			  "j_mount_id %ld\n", 
			  bh->b_blocknr,
			  j_head->j_last_flush_trans_id,
			  j_head->j_first_unflushed_offset,
			  j_head->j_mount_id);
	brelse (bh);
    }
}
#endif
